from rest_framework import serializers

from referrals.models import (
    PayoutMethod,
    Promoter,
    PromoterCommission,
    PromoterPayout,
    Referral, ReferralProgram,
)
from referrals.repositories.promoter_commission_repository import promoter_commission_repository


class CamelCaseSerializer(serializers.ModelSerializer):

    def get_current_user(self):
        user = None
        request = self.context.get("request")
        if request and hasattr(request, "user"):
            user = request.user
        return user

    def to_representation(self, instance):
        ret = super().to_representation(instance)
        return {snake_case_to_camel_case(key): value for key, value in ret.items()}


def snake_case_to_camel_case(snake_case):
    components = snake_case.split("_")
    return components[0] + "".join(x.title() for x in components[1:])


class PromoterCommissionSerializer(CamelCaseSerializer):
    class Meta:
        model = PromoterCommission
        fields = "__all__"


class ReferralSerializer(CamelCaseSerializer):
    email = serializers.SerializerMethodField()
    user_id = serializers.SerializerMethodField()
    commission_amount = serializers.SerializerMethodField()
    commission_status = serializers.SerializerMethodField()

    class Meta:
        model = Referral
        fields = (
            "user_id",
            "email",
            "status",
            "invitation_method",
            "commission_rate",
            "commission_amount",
            "commission_status",
        )

    def get_email(self, obj):
        return obj.user.email

    def get_user_id(self, obj):
        return obj.user.id

    def get_commission_amount(self, obj):
        commission = promoter_commission_repository.get_referral_positive_commission(obj)
        if commission:
            return commission.amount
        return 0

    def get_commission_status(self, obj):
        commission = promoter_commission_repository.get_referral_positive_commission(obj)
        if commission:
            return commission.status
        return None


class PayoutMethodSerializer(CamelCaseSerializer):
    class Meta:
        model = PayoutMethod
        fields = ["method", "payment_address"]


class PromoterSerializer(CamelCaseSerializer):
    active_payout_method = PayoutMethodSerializer()
    commission_rate = serializers.SerializerMethodField()

    class Meta:
        model = Promoter
        fields = (
            "id",
            "user",
            "referral_token",
            "referral_link",
            "active_payout_method",
            "current_balance",
            "total_earned",
            "total_paid",
            "created",
            "updated",
            "link_clicked",
            "min_withdrawal_balance",
            "commission_rate",
        )

    def get_commission_rate(self, obj):
        active_program = ReferralProgram.get_active_referral_program()
        return active_program.commission_rate


class PromoterPayoutsSerializer(CamelCaseSerializer):
    class Meta:
        model = PromoterPayout
        fields = ["id", "created", "amount"]


class MinWithdrawalBalanceSerializer(PromoterSerializer):
    class Meta(PromoterSerializer.Meta):
        fields = ['min_withdrawal_balance']



import logging

from django.contrib.auth.models import User

from referrals.config import config
from referrals.models import Promoter
from referrals.repositories.promoter_repository import promoter_repository
from referrals.services.referral_service import referral_service

logger = logging.getLogger(__name__)


class PromoterService:
    """
    Service class for managing Promoter-related operations.
    """

    def create_new_promoter(self, user: User) -> Promoter:
        referral_token = referral_service.generate_referral_token(user_id=user.id)
        referral_link = referral_service.generate_referral_link(base_referral_link=config.BASE_REFERRAL_LINK,
                                                                referral_token=referral_token)

        promoter = Promoter(
            user=user,
            referral_token=referral_token,
            referral_link=referral_link,
        )
        promoter.save()
        logger.info(f"Created new promoter {user.email}")

        return promoter

    def get_or_create_promoter(self, user: User) -> Promoter:
        promoter = promoter_repository.get_by_user_id(user.id)
        if not promoter:
            promoter = self.create_new_promoter(user=user)

        return promoter


promoter_service = PromoterService()



from typing import Union


class BaseAPIException(Exception):
    """
    Base API exception.
    """

    def __init__(self, error_message: Union[str, dict], status_code: int):
        self.error_message = error_message
        self.status_code = status_code
        super().__init__(self.error_message)


class ViewException(BaseAPIException):
    """Base view exception."""



from typing import Generic, List, Optional, Tuple, Type, TypeVar

from django.db.models import Model, QuerySet
from django.shortcuts import get_object_or_404

T = TypeVar("T", bound=Model)


class BaseRepository(Generic[T]):
    def __init__(self, model: Type[T]):
        self.model = model

    def get_one(self, **kwargs) -> Optional[T]:
        try:
            return self.model.objects.get(**kwargs)
        except self.model.DoesNotExist:
            return None

    def get_object_or_404(self, **kwargs) -> T:
        return get_object_or_404(self.model, **kwargs)

    def get_all(self) -> QuerySet[T]:
        return self.model.objects.all()

    def get_or_create(self, defaults: Optional[dict] = None, **kwargs) -> Tuple[T, bool]:
        return self.model.objects.get_or_create(defaults=defaults, **kwargs)

    def create(self, **kwargs) -> T:
        return self.model.objects.create(**kwargs)

    def create_many(self, data_list: List[dict]) -> List[T]:
        instances = [self.model(**data) for data in data_list]
        return self.model.objects.bulk_create(instances)

    def update(self, values: dict, **kwargs) -> T:
        instance = self.model.objects.get(**kwargs)
        for attr, value in values.items():
            setattr(instance, attr, value)
        instance.save()
        return instance

    def update_or_create(self, defaults: Optional[dict] = None, **kwargs) -> Tuple[T, bool]:
        return self.model.objects.update_or_create(defaults=defaults, **kwargs)

    def filter(self, **kwargs) -> QuerySet[T]:
        return self.model.objects.filter(**kwargs)

    def exclude(self, **kwargs) -> QuerySet[T]:
        return self.model.objects.exclude(**kwargs)

    def select_for_update(self) -> QuerySet[T]:
        return self.model.objects.select_for_update()

    def filter_one(self, **kwargs) -> Optional[T]:
        return self.model.objects.filter(**kwargs).first()

    def delete(self, db_obj: Optional[T] = None, **kwargs) -> bool:
        if db_obj:
            deleted_count, _ = db_obj.delete()
            return deleted_count > 0

        instances = self.filter(**kwargs)
        deleted_count, _ = instances.delete()
        return deleted_count > 0

    def select_related(self, *args) -> QuerySet[T]:
        return self.model.objects.select_related(*args)

    def prefetch_related(self, *args) -> QuerySet[T]:
        return self.model.objects.prefetch_related(*args)

    def bulk_create(self, objs: List[T]) -> List[T]:
        return self.model.objects.bulk_create(objs)

    def bulk_update(self, objs: List[T], fields: Optional[List[str]] = None) -> None:
        return self.model.objects.bulk_update(objs, fields)

    def values_list(self, *fields, flat: bool = False, named: bool = False) -> QuerySet:
        return self.model.objects.values_list(*fields, flat=flat, named=named)



import math
from datetime import datetime, timedelta
from decimal import Decimal

from django.contrib.auth.models import User
from django.test import TestCase
from django.urls import reverse
from rest_framework import status
from rest_framework.test import APITestCase, APIClient

from referrals.choices import InvitationMethodChoices, ReferralStateChoices, PromoterCommissionStatusChoices
from referrals.config import config
from referrals.exceptions import ViewException
from referrals.models import ReferralProgram, Promoter, Referral, PromoterPayout, PromoterCommission
from referrals.serializers import ReferralSerializer, PromoterSerializer, PromoterPayoutsSerializer
from referrals.services import referral_service, promoter_service
from referrals.services.promoter_payout_service import promoter_payout_service


class ReferralProgramViewSetTestCase(APITestCase):
    def setUp(self):
        self.referral_program = ReferralProgram.objects.create(name='test_program', commission_rate=20.00,
                                                               is_active=True, min_withdrawal_balance=10)
        self.user = User.objects.create_user(username='test-user', email='test@example.com', password='Password123')
        self.user2 = User.objects.create_user(username='test-user2', email='test2@example.com', password='Password321')
        self.user3 = User.objects.create_user(username='test-user3', email='test3@example.com', password='Password121')
        self.user4 = User.objects.create_user(username='test-user4', email='test4@example.com', password='Password111')

        self.promoter = Promoter.objects.create(user=self.user, referral_token='test-token')

        self.referral = Referral.objects.create(
            user=self.user2,
            promoter=self.promoter,
            status=ReferralStateChoices.ACTIVE
        )
        self.referral2 = Referral.objects.create(
            user=self.user3,
            promoter=self.promoter,
            status=ReferralStateChoices.ACTIVE
        )
        self.client = APIClient()
        self.client.force_authenticate(user=self.user)

    def test_create_referral_success(self):
        url = reverse('referrals-list')
        data = {
            "email": self.user4.email,
            "referral_token": "test-token",
            "referral_source": InvitationMethodChoices.EMAIL.value,
        }

        response = self.client.post(url, data, format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertTrue(Referral.objects.filter(user=self.user4).exists())

        referral = Referral.objects.get(user=self.user4)
        self.assertEqual(referral.promoter, self.promoter)
        self.assertEqual(referral.invitation_method, InvitationMethodChoices.EMAIL.value)
        self.assertEqual(referral.status, ReferralStateChoices.SIGNUP.value)

    def test_create_referral_self_referral(self):
        url = reverse('referrals-list')
        data = {
            "email": self.user.email,
            "referral_token": "test-token",
            "referral_source": InvitationMethodChoices.LINK.value,
        }

        with self.assertRaises(ViewException) as context:
            self.client.post(url, data, format='json')

        self.assertIn("You can't refer to yourself.", str(context.exception))

    def test_create_referral_invalid_token(self):
        url = reverse('referrals-list')
        data = {
            "email": self.user4.email,
            "referral_token": "invalid-token",  # Invalid token
            "referral_source": InvitationMethodChoices.LINK.value,
        }

        response = self.client.post(url, data, format='json')

        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertFalse(Referral.objects.filter(user=self.user4).exists())

    def test_list_referrals(self):
        url = reverse('referrals-list')
        response = self.client.get(url)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIn("results", response.data)
        self.assertIsInstance(response.data["results"], list)

        expected_data = ReferralSerializer([self.referral, self.referral2], many=True).data

        response_data_sorted = sorted(response.data["results"], key=lambda x: x['userId'])
        expected_data_sorted = sorted(expected_data, key=lambda x: x['userId'])

        self.assertEqual(response_data_sorted, expected_data_sorted)

    def test_list_referrals_empty(self):
        Referral.objects.all().delete()

        url = reverse('referrals-list')
        response = self.client.get(url)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIn("results", response.data)
        self.assertEqual(response.data["results"], [])

    def test_get_referral_link(self):
        self.client.force_authenticate(user=self.user2)

        url = reverse('referrals-get-referral-link')
        response = self.client.get(url)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertTrue(Promoter.objects.filter(user=self.user2).exists())

        self.assertIn('referralLink', response.data)
        self.assertEqual(response.data['referralLink'], self.user2.promoter.referral_link)

    def test_retrieve_promoter(self):
        url = reverse('referrals-retrieve-promoter')
        response = self.client.get(url)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        promoter = Promoter.objects.get(user=self.user)
        serializer = PromoterSerializer(promoter)

        self.assertEqual(response.data, serializer.data)

    def test_set_payout_method_to_promoter(self):
        url = reverse('referrals-set-payout-method')
        data = {
            "method": "wise",
            "payment_address": "test@example.com"
        }
        response = self.client.patch(url, data, format='json')

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.promoter.refresh_from_db()
        self.assertEqual(self.promoter.active_payout_method.method, "wise")
        self.assertEqual(self.promoter.active_payout_method.payment_address, "test@example.com")

    def test_set_min_withdrawal_balance_success(self):
        url = reverse('referrals-set-min-withdrawal-balance')
        data = {
            "min_withdrawal_balance": 15.00
        }
        response = self.client.patch(url, data, format='json')

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.promoter.refresh_from_db()
        self.assertEqual(self.promoter.min_withdrawal_balance, 15.00)

        expected_data = PromoterSerializer(self.promoter).data
        self.assertEqual(response.data, expected_data)

    def test_set_min_withdrawal_balance_below_program_min(self):
        url = reverse('referrals-set-min-withdrawal-balance')
        data = {
            "min_withdrawal_balance": 5.00
        }

        with self.assertRaises(ViewException) as context:
            self.client.patch(url, data, format='json')

        self.assertIn(
            "Min withdrawal balance must be greater than or equal to the referral program's min withdrawal balance",
            str(context.exception))

    def test_increment_link_clicked_success(self):
        url = reverse('referrals-increment-link-clicked')
        data = {
            "referral_token": "test-token"
        }

        response = self.client.post(url, data, format='json')

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIn("message", response.data)
        self.assertEqual(response.data["message"], "Link clicked count incremented successfully")

        self.promoter.refresh_from_db()
        self.assertEqual(self.promoter.link_clicked, 1)

    def test_increment_link_clicked_invalid_token(self):
        url = reverse('referrals-increment-link-clicked')
        data = {
            "referral_token": "invalid-token"
        }

        response = self.client.post(url, data, format='json')

        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_promoter_payment_history_with_payouts(self):
        payout1 = PromoterPayout.objects.create(
            promoter=self.promoter, amount=100, payout_method="wise", tx_signature="tx123"
        )
        payout2 = PromoterPayout.objects.create(
            promoter=self.promoter, amount=200, payout_method="crypto", tx_signature="tx456"
        )

        url = reverse('referrals-promoter-payment-history')
        response = self.client.get(url)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIsInstance(response.data, list)

        expected_data = PromoterPayoutsSerializer([payout2, payout1], many=True).data
        self.assertEqual(response.data, expected_data)

    def test_promoter_payment_history_no_payouts(self):
        url = reverse('referrals-promoter-payment-history')
        response = self.client.get(url)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data, [])

    def test_promoter_recent_earnings(self):
        seven_days_ago = datetime.today().date() - timedelta(days=6)
        PromoterCommission.objects.create(promoter=self.promoter, amount=100, referral=self.referral,
                                          created=seven_days_ago)
        PromoterCommission.objects.create(promoter=self.promoter, amount=200, referral=self.referral2,
                                          created=datetime.today())

        url = reverse('referrals-promoter-recent-earnings')
        response = self.client.get(url)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIsInstance(response.data, list)
        self.assertEqual(len(response.data), 7)

        total_earnings = sum(item['value'] for item in response.data)
        self.assertEqual(total_earnings, 300)

    def test_promoter_recent_earnings_no_data(self):
        url = reverse('referrals-promoter-recent-earnings')
        response = self.client.get(url)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIsInstance(response.data, list)
        self.assertEqual(len(response.data), 7)

        # Check that all values are 0 when there's no data
        for day_data in response.data:
            self.assertEqual(day_data['value'], 0)

    def test_promoter_payment_history(self):
        payout1 = PromoterPayout.objects.create(
            promoter=self.promoter, amount=100, payout_method="wise", tx_signature="tx123"
        )
        payout2 = PromoterPayout.objects.create(
            promoter=self.promoter, amount=200, payout_method="crypto", tx_signature="tx456"
        )

        url = reverse('referrals-promoter-payment-history')
        response = self.client.get(url)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIsInstance(response.data, list)
        self.assertEqual(len(response.data), 2)

        self.assertEqual(response.data[0]['amount'], 200)
        self.assertEqual(response.data[1]['amount'], 100)

    def test_promoter_payment_history_no_data(self):
        url = reverse('referrals-promoter-payment-history')
        response = self.client.get(url)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIsInstance(response.data, list)
        self.assertEqual(len(response.data), 0)

    def tearDown(self):
        ReferralProgram.objects.all().delete()
        Promoter.objects.all().delete()
        Referral.objects.all().delete()
        PromoterPayout.objects.all().delete()
        PromoterCommission.objects.all().delete()
        User.objects.all().delete()


class ReferralServiceTestCase(TestCase):
    commission_rate = 20.00

    def setUp(self):
        self.referral_program = ReferralProgram.objects.create(name='test_program',
                                                               commission_rate=self.commission_rate,
                                                               is_active=True, min_withdrawal_balance=10)
        self.user = User.objects.create_user(username='test-user', email='test@example.com', password='Password123')
        self.user2 = User.objects.create_user(username='test-user2', email='test2@example.com', password='Password321')
        self.user3 = User.objects.create_user(username='test-user3', email='test3@example.com', password='Password121')

        self.promoter = Promoter.objects.create(user=self.user, referral_token='test-token')
        self.referral = Referral.objects.create(
            user=self.user2,
            promoter=self.promoter,
            status=ReferralStateChoices.SIGNUP
        )
        self.referral2 = Referral.objects.create(
            user=self.user3,
            promoter=self.promoter,
            status=ReferralStateChoices.SIGNUP
        )
        self.client = APIClient()
        self.client.force_authenticate(user=self.user)

    def test_get_user_earnings(self):
        seven_days_ago = datetime.today().date() - timedelta(days=6)
        PromoterCommission.objects.create(promoter=self.promoter, referral=self.referral, amount=100,
                                          created=seven_days_ago)
        PromoterCommission.objects.create(promoter=self.promoter, referral=self.referral2, amount=200,
                                          created=datetime.today())

        earnings = referral_service.get_user_earnings(self.user)
        self.assertEqual(len(earnings), 2)
        self.assertEqual(earnings[0]['amount'], 100)
        self.assertEqual(earnings[1]['amount'], 200)

    def test_aggregate_earnings_by_day(self):
        earnings = [
            {"created": datetime.today().strftime("%Y-%m-%dT%H:%M:%S.%fZ"), "amount": 100},
            {"created": (datetime.today() - timedelta(days=1)).strftime("%Y-%m-%dT%H:%M:%S.%fZ"), "amount": 200},
        ]
        aggregated_earnings = referral_service.aggregate_earnings_by_day(earnings)
        today = datetime.today().strftime("%a")
        yesterday = (datetime.today() - timedelta(days=1)).strftime("%a")

        self.assertEqual(aggregated_earnings[today], 100)
        self.assertEqual(aggregated_earnings[yesterday], 200)

    def test_generate_referral_token(self):
        referral_token = referral_service.generate_referral_token(self.user.id)
        self.assertEqual(len(referral_token), 10)

    def test_get_referrer_by_user_id(self):
        referrer = referral_service.get_referrer_by_user_id(self.user2.id)
        self.assertEqual(referrer, self.promoter)

    def test_handle_purchase_subscription(self):
        self.referral.status = ReferralStateChoices.SIGNUP
        self.referral.save()

        amount_paid = 15000  # amount in cents

        commission = referral_service.handle_purchase_subscription(self.user2, amount_paid)
        self.referral.refresh_from_db()

        self.assertEqual(self.referral.status, ReferralStateChoices.ACTIVE)
        expected_commission_amount = Decimal(amount_paid * (self.commission_rate / 100) / 100)

        self.assertIsNotNone(commission)
        self.assertEqual(commission.amount, expected_commission_amount)
        self.assertEqual(commission.promoter, self.promoter)
        self.assertEqual(commission.referral, self.referral)

    def test_handle_user_refund(self):
        self.referral.status = ReferralStateChoices.ACTIVE
        self.referral.save()

        amount_paid = 15000  # amount in cents
        amount_refunded = 5000  # amount refunded in cents

        initial_commission = promoter_payout_service.create_commission(
            referral=self.referral, amount_paid=amount_paid
        )

        result = referral_service.handle_user_refund(self.user2, amount_refunded=amount_refunded,
                                                     amount_paid=amount_paid)
        self.referral.refresh_from_db()

        expected_refund_amount = -math.floor(initial_commission.amount * amount_refunded / amount_paid)

        self.assertEqual(result.amount, expected_refund_amount)
        self.assertEqual(result.status, PromoterCommissionStatusChoices.REFUND)
        self.assertEqual(self.referral.status, ReferralStateChoices.REFUND)

    def tearDown(self):
        User.objects.all().delete()
        ReferralProgram.objects.all().delete()
        Promoter.objects.all().delete()
        Referral.objects.all().delete()
        PromoterCommission.objects.all().delete()


class PromoterServiceTestCase(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user = User.objects.create_user(username='test-user', email='test@example.com', password='Password123')

    def test_create_new_promoter(self):
        promoter = promoter_service.create_new_promoter(user=self.user)

        self.assertIsNotNone(promoter)
        self.assertEqual(promoter.user, self.user)
        self.assertTrue(len(promoter.referral_token) > 0)
        self.assertTrue(promoter.referral_link.startswith(config.BASE_REFERRAL_LINK))

        self.assertTrue(Promoter.objects.filter(user=self.user).exists())

    def test_get_or_create_promoter_existing(self):
        existing_promoter = Promoter.objects.create(
            user=self.user,
            referral_token='existing-token',
            referral_link='http://example.com/referral?ref=existingtoken'
        )

        promoter = promoter_service.get_or_create_promoter(user=self.user)

        self.assertEqual(promoter, existing_promoter)
        self.assertTrue(Promoter.objects.filter(user=self.user).exists())

    @classmethod
    def tearDownClass(cls):
        Promoter.objects.all().delete()
        User.objects.all().delete()



from typing import Optional

from .base_repository import BaseRepository
from referrals.models import Referral


class ReferralRepository(BaseRepository):

    def get_referrals_by_user_id(self, user_id: int) -> Optional[Referral]:
        return self.select_related("promoter__user").filter(promoter__user_id=user_id).all().order_by("-created")

    def get_referral_by_user_id(self, user_id: int) -> Optional[Referral]:
        return self.select_related("promoter").filter(user_id=user_id).first()


referral_repository = ReferralRepository(model=Referral)



from django.db import models


class ReferralStateChoices(models.TextChoices):
    SIGNUP = "signup"
    ACTIVE = "active"
    REFUND = "refund"


class InvitationMethodChoices(models.TextChoices):
    EMAIL = "email"
    LINK = "link"


class PromoterCommissionStatusChoices(models.TextChoices):
    PENDING = "pending"
    PAID = "paid"
    FAILED = "failed"
    REFUND = "refund"



import logging

from django.contrib.auth.models import User
from rest_framework import permissions, status, viewsets
from rest_framework.decorators import action
from rest_framework.pagination import PageNumberPagination
from rest_framework.response import Response
from rest_framework.status import HTTP_200_OK, HTTP_201_CREATED

from referrals.choices import (
    InvitationMethodChoices,
    ReferralStateChoices,
)
from referrals.exceptions import ViewException
from referrals.models import PayoutMethod, PromoterPayout, ReferralProgram
from referrals.repositories.promoter_repository import promoter_repository
from referrals.repositories.referral_repository import referral_repository
from referrals.serializers import (
    PayoutMethodSerializer,
    PromoterPayoutsSerializer,
    PromoterSerializer,
    ReferralSerializer, MinWithdrawalBalanceSerializer,
)
from referrals.services import promoter_service, referral_service

logger = logging.getLogger(__name__)


class ReferralsPagination(PageNumberPagination):
    page_size = 10
    page_size_query_param = "page_size"

    def get_paginated_response(self, data):
        return Response(
            {
                "count": self.page.paginator.count,
                "pages": self.page.paginator.num_pages,
                "results": data,
            }
        )


class ReferralProgramViewSet(
    viewsets.GenericViewSet,
):
    pagination_class = ReferralsPagination
    serializer_class = ReferralSerializer

    def get_permissions(self):
        allow_any_actions = ["create", "increment_link_clicked"]
        if self.action in allow_any_actions:
            return [permissions.AllowAny()]
        return super().get_permissions()

    def get_queryset(self):
        return referral_repository.get_referrals_by_user_id(self.request.user.id)

    @action(detail=False, methods=["GET"], url_path="get-referral-link")
    def get_referral_link(self, request, *args, **kwargs):
        user = request.user
        promoter = promoter_service.get_or_create_promoter(user=user)
        return Response({"referralLink": promoter.referral_link}, status=HTTP_200_OK)

    @action(detail=False, methods=["GET"], url_path="promoter")
    def retrieve_promoter(self, request, *args, **kwargs):
        user = request.user
        promoter = promoter_service.get_or_create_promoter(user=user)
        serializer = PromoterSerializer(promoter)
        return Response(serializer.data)

    @action(detail=False, methods=["PATCH"], url_path="set-payout-method")
    def set_payout_method(self, request, *args, **kwargs):
        promoter = request.user.promoter

        serializer = PayoutMethodSerializer(data=request.data)
        if not serializer.is_valid():
            return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)

        method = serializer.validated_data.get("method")
        payment_address = serializer.validated_data.get("payment_address")

        if promoter.active_payout_method:
            promoter.active_payout_method.method = method
            promoter.active_payout_method.payment_address = payment_address
            promoter.active_payout_method.save()
        else:
            payout_method = PayoutMethod.objects.create(method=method, payment_address=payment_address)
            promoter.active_payout_method = payout_method
            promoter.save()

        promoter_serializer = PromoterSerializer(promoter)

        return Response(promoter_serializer.data, status=status.HTTP_200_OK)

    @action(detail=False, methods=["PATCH"], url_path="set-min-withdrawal-balance")
    def set_min_withdrawal_balance(self, request, *args, **kwargs):
        promoter = request.user.promoter

        serializer = MinWithdrawalBalanceSerializer(data=request.data)
        if not serializer.is_valid():
            return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)

        min_withdrawal_balance = serializer.validated_data.get("min_withdrawal_balance")
        program_min_withdrawal_balance = ReferralProgram.get_active_referral_program().min_withdrawal_balance

        if min_withdrawal_balance < program_min_withdrawal_balance:
            raise ViewException(
                f"Min withdrawal balance must be greater than or equal to the referral program's min withdrawal balance ({program_min_withdrawal_balance}).",
                status_code=400)

        promoter.min_withdrawal_balance = min_withdrawal_balance
        promoter.save()

        promoter_serializer = PromoterSerializer(promoter)
        return Response(promoter_serializer.data, status=status.HTTP_200_OK)

    @action(detail=False, methods=["GET"], url_path="promoter-recent-earnings")
    def promoter_recent_earnings(self, request, *args, **kwargs):
        user = request.user

        earnings = referral_service.get_user_earnings(user)
        result = referral_service.get_last_7_days_earnings(earnings)
        return Response(result, status=HTTP_200_OK)

    @action(detail=False, methods=["GET"], url_path="payouts")
    def promoter_payment_history(self, request, *args, **kwargs):
        user = request.user

        payouts = PromoterPayout.objects.filter(promoter__user=user).order_by("-created")
        serializer = PromoterPayoutsSerializer(payouts, many=True)
        return Response(serializer.data)

    def list(self, request, *args, **kwargs):
        """List referrals objects"""
        queryset = self.get_queryset()
        page = self.paginate_queryset(queryset)
        user = request.user
        promoter_service.get_or_create_promoter(user=user)

        serializer = self.get_serializer(page, many=True)
        response_data = self.get_paginated_response(serializer.data).data
        response_data.setdefault("results", [])

        return Response(response_data, status=HTTP_200_OK)

    def create(self, request, *args, **kwargs):
        """Creates a new referral object."""
        user = User.objects.get(email=request.data["email"])

        referral_token: str = request.data.get("referral_token")
        invitation_method: str = request.data.get("referral_source", InvitationMethodChoices.LINK.value)
        promoter = promoter_repository.get_object_or_404(referral_token=referral_token)

        if promoter.user == user:
            raise ViewException("You can't refer to yourself.", status_code=400)

        referral = referral_repository.create(
            user=user,
            promoter=promoter,
            invitation_method=invitation_method if invitation_method else InvitationMethodChoices.LINK.value,
            status=ReferralStateChoices.SIGNUP.value,
        )
        serializer = self.get_serializer(referral)
        return Response(serializer.data, status=HTTP_201_CREATED)

    @action(detail=False, methods=["POST"], url_path="increment-link-clicked")
    def increment_link_clicked(self, request, *args, **kwargs):
        referral_token = request.data.get("referral_token")
        promoter = promoter_repository.get_object_or_404(referral_token=referral_token)

        promoter.link_clicked = promoter.link_clicked + 1
        promoter.save()
        return Response({"message": "Link clicked count incremented successfully"}, status=status.HTTP_200_OK)



import hashlib
import logging
from collections import defaultdict
from datetime import datetime, timedelta
from typing import Optional

from django.contrib.auth.models import User
from django.core.exceptions import ObjectDoesNotExist
from django.core.mail import EmailMessage
from django.db import transaction
from django.template.exceptions import TemplateDoesNotExist
from django.template.loader import get_template

from referrals.choices import ReferralStateChoices
from referrals.config import config
from referrals.models import PromoterCommission, Promoter
from referrals.serializers import PromoterCommissionSerializer
from referrals.services.promoter_payout_service import promoter_payout_service
from referrals.utils import append_query_params

logger = logging.getLogger(__name__)


class ReferralService:
    """
    Service class that handles operations related to referral programs, including sending invitation emails,
    managing user earnings, generating referral tokens and links, and handling subscriptions and refunds.
    """

    @staticmethod
    def send_referral_invitation_email(emails_to: list[str], invitation_link: str,
                                       promoter_full_name: str, subject: str, template_path: str,
                                       from_email: str = config.BASE_EMAIL) -> bool:
        """
        Sends an HTML email with an invitation link to the specified email addresses.

        :param emails_to: List of email addresses to send the invitation to.
        :param invitation_link: The referral invitation link.
        :param promoter_full_name: Full name of the promoter.
        :param subject: Subject of the email.
        :param template_path: Path to the HTML template for the email.
        :param from_email: The email address that will appear in the 'from' field.

        :return: True if the email was sent successfully, False otherwise.
        """
        invitation_link = append_query_params(invitation_link, {"ref-source": "email"})
        html_template_context = {
            "link": invitation_link,
            "promoter_full_name": promoter_full_name,
        }

        try:
            template = get_template(template_path)
            email_message = template.render(html_template_context)

            email = EmailMessage(
                subject,
                body=email_message,
                from_email=from_email,
                to=emails_to,
            )
            email.content_subtype = "html"
            email.send()
            logging.info(f"Referral invitation email sent to {emails_to}")
            return True
        except TemplateDoesNotExist:
            logging.error(f"Template '{template_path}' does not exist.")
            return False

    @staticmethod
    def get_user_earnings(user: User):
        """
        Retrieves the earnings of a user within the last 7 days.

        Filters the commissions associated with the user's promoter account that were created within
        the last 7 days, serializes the data, and returns it.

        Args:
            user (User): The user whose earnings are to be retrieved.

        Returns:
            list[dict]: A list of serialized commission data for the user's earnings.
        """
        seven_days_ago = datetime.today().date() - timedelta(days=6)
        earnings = PromoterCommission.objects.filter(promoter__user=user, created__gte=seven_days_ago)
        serializer = PromoterCommissionSerializer(earnings, many=True)
        return serializer.data

    @staticmethod
    def aggregate_earnings_by_day(earnings: list[dict]) -> dict[str, int]:
        """
        Aggregates earnings by the day of the week.

        Processes a list of earnings, grouping the total earnings by the day of the week they were created.

        Args:
            earnings (list[dict]): A list of earnings with their creation dates.

        Returns:
            dict[str, int]: A dictionary mapping each day of the week to the total earnings for that day.
        """
        earnings_by_day = defaultdict(int)

        for earning in earnings:
            created_at = datetime.strptime(earning["created"], "%Y-%m-%dT%H:%M:%S.%fZ")
            day_of_week = created_at.strftime("%a")
            earnings_by_day[day_of_week] += earning["amount"]
        return earnings_by_day

    @staticmethod
    def get_last_7_days_earnings(earnings: list[dict]) -> list[dict]:
        """
        Retrieves the earnings statistics for the last 7 days.

        Aggregates earnings data by the day of the week for the last 7 days, ensuring each day is represented
        even if there were no earnings on that day.

        Args:
            earnings (list[dict]): A list of earnings with their creation dates.

        Returns:
            list[dict]: A list of dictionaries, each containing the day (as a string) and the corresponding earnings value.
        """
        today = datetime.today()
        last_7_days = [(today - timedelta(days=i)).strftime("%a") for i in range(6, -1, -1)]
        earnings_by_day = ReferralService.aggregate_earnings_by_day(earnings)
        statistics = [{"day": day[:2], "value": earnings_by_day.get(day, 0)} for day in last_7_days]
        return statistics

    @staticmethod
    def generate_referral_token(user_id: int) -> str:
        """
        Generate a unique referral code using the user's ID and the current timestamp,
        hashed for consistency in length and to obscure the user ID.

        :param user_id: The ID of the user from whom the referral code is being generated.
        :return: A unique referral code.
        """

        raw_code = f"{user_id}"
        hash_object = hashlib.sha256(raw_code.encode())
        referral_code = hash_object.hexdigest()[:10]

        return referral_code.upper()

    @staticmethod
    def generate_referral_link(base_referral_link: str, referral_token: str) -> str:
        """
        Generate a unique referral link with a referral token.
        """
        return append_query_params(base_referral_link, {"ref": referral_token})

    @staticmethod
    def get_referrer_by_user_id(user_id: int) -> Optional[Promoter]:
        """
        Retrieve the promoter (referrer) associated with a given user ID.
        """
        try:
            user = User.objects.select_related("referral__promoter__user").filter(pk=user_id).first()
            return user.referral.promoter
        except User.DoesNotExist:
            return None
        except ObjectDoesNotExist:
            logger.error("User does not have referral relation")
            return None

    @staticmethod
    def handle_purchase_subscription(user: User,
                                     amount_paid: int,
                                     invoice_external_id: Optional[int] = None) -> Optional[PromoterCommission]:
        """
        Handles the process of updating a referral subscription status to 'Active'
        when a subscription is created for a referred user.

        Args:
            user (User): The user whose subscription is being handled.
            amount_paid (int): The amount paid for the subscription in cents.
            invoice_external_id (Optional[int]): An external ID for the invoice (optional).

        Returns:
            Optional[PromoterCommission]: The commission generated from the subscription, or None if no commission is generated.
        """
        try:
            user = User.objects.select_related("referral__promoter__user").filter(pk=user.id).first()
            promoter = user.referral.promoter
            if user.referral.status == ReferralStateChoices.SIGNUP:
                user.referral.status = ReferralStateChoices.ACTIVE
                user.referral.save()
                commission = promoter_payout_service.calculate_commission(user.id, amount_paid, invoice_external_id)
                logger.info(f"User {user.email} became an active referral of {promoter.user.email}")
                return commission
        except ObjectDoesNotExist:
            logger.error(f"User with ID {user.id} has no referral associated.")

    @staticmethod
    @transaction.atomic
    def handle_user_refund(user: User, amount_refunded: int, amount_paid: int,
                           invoice_external_id: Optional[int] = None) -> Optional[PromoterCommission]:
        """
        Handles the process of refunding a referred user's subscription.

        This method updates the user's referral status to 'Refund' and calculates the refund commission.

        Args:
            user (User): The user whose refund is being processed.
            amount_refunded (int): The amount refunded in cents.
            amount_paid (int): The original amount paid for the subscription in cents.
            invoice_external_id (Optional[int]): An external ID for the invoice (optional).

        Returns:
            Optional[PromoterCommission]: The refund commission generated from the refund, or None if no commission is generated.
        """
        try:
            user = User.objects.select_related("referral__promoter__user").filter(pk=user.id).first()

            if user.referral.status == ReferralStateChoices.ACTIVE:
                user.referral.status = ReferralStateChoices.REFUND
                user.referral.save()
                commission = promoter_payout_service.calculate_refund(user.referral, amount_refunded, amount_paid,
                                                                      invoice_external_id)
                logger.info(f"User {user.email} has been refunded {amount_refunded}.")
                return commission
        except ObjectDoesNotExist:
            logger.error(f"User with ID {user.id} has no referral associated.")


referral_service = ReferralService()



from enum import Enum


class ReferralStateEnum(Enum):
    SIGNUP = "signup"
    INVITED = "invited"
    ACTIVE = "active"


class PayoutStatusEnum(Enum):
    PENDING = "pending"
    COMPLETED = "completed"
    PROCESSING = "processing"


class CryptoPayoutTokenIdsEnum(Enum):
    USDC = "usd-coin"



from rest_framework.routers import SimpleRouter

from referrals.views import ReferralProgramViewSet

router = SimpleRouter()
router.register("", ReferralProgramViewSet, basename="referrals")

urlpatterns = []

urlpatterns += router.urls



import logging
from typing import Optional

from referrals.models import Promoter, Referral
from .base_repository import BaseRepository

logger = logging.getLogger(__name__)


class PromoterRepository(BaseRepository):
    def get_by_user_id(self, user_id: int) -> Optional[Promoter]:
        return self.select_related("user").filter(user_id=user_id).first()

    def get_by_referral_token(self, referral_token: str) -> Optional[Promoter]:
        return self.select_related("user").filter(referral_token=referral_token).first()

    def get_wise_payout_promoters(self):
        return (
            self.select_related("user", "active_payout_method").filter(
                active_payout_method__method="wise")
        )

    def check_promoter_get_commission_from_referral(self, promoter: Promoter, referral: Referral) -> bool:
        return self.filter(promoter_commission__referral=referral, pk=promoter.id).exists()


promoter_repository = PromoterRepository(model=Promoter)



from django.contrib import admin

from referrals.models import ReferralProgram, PayoutMethod, Referral, Promoter, PromoterCommission, \
    PromoterPayout


@admin.register(ReferralProgram)
class ReferralProgramAdmin(admin.ModelAdmin):
    list_display = ("name", "commission_rate", "is_active", "id")


@admin.register(PayoutMethod)
class PayoutMethodAdmin(admin.ModelAdmin):
    list_display = ("method", "payment_address")


@admin.register(Referral)
class ReferralAdmin(admin.ModelAdmin):
    autocomplete_fields = (
        "user",
        "promoter",
    )
    list_display = (
        "id",
        "user",
        "promoter",
        "invitation_method",
        "status",
    )
    search_fields = ("id",)
    list_filter = ("status", "invitation_method", "id")


@admin.register(Promoter)
class PromoterAdmin(admin.ModelAdmin):
    autocomplete_fields = ("user",)
    search_fields = ("user__email",)
    list_display = ("referral_link", "referral_token", "user", "id")


@admin.register(PromoterCommission)
class PromoterCommissionAdmin(admin.ModelAdmin):
    search_fields = ("promoter__user__email",)
    list_display = ("promoter", "referral", "amount", "status")
    list_filter = ("status", "amount")


@admin.register(PromoterPayout)
class PromoterPayoutsAdmin(admin.ModelAdmin):
    search_fields = ("promoter_user_email",)
    list_display = ("promoter", "amount", "payout_method", "created")
    list_filter = ("payout_method",)



# Generated by Django 5.1 on 2024-08-17 19:32

import django.core.validators
import django.db.models.deletion
from decimal import Decimal
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='PayoutMethod',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('method', models.CharField(choices=[('wise', 'Wise'), ('crypto', 'Crypto')], max_length=10)),
                ('payment_address', models.CharField(help_text='Payment address (email or solana wallet)', max_length=100)),
            ],
        ),
        migrations.CreateModel(
            name='ReferralProgram',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True)),
                ('name', models.CharField(max_length=255, unique=True)),
                ('commission_rate', models.DecimalField(decimal_places=2, help_text='Commission rate as a percentage', max_digits=5, validators=[django.core.validators.MinValueValidator(Decimal('0.01'))])),
                ('is_active', models.BooleanField(default=True, help_text='Indicates if this is the currently active referral program')),
                ('min_withdrawal_balance', models.DecimalField(decimal_places=2, default=0.0, help_text='Minimum balance required to withdraw earnings', max_digits=10)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Promoter',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True)),
                ('referral_token', models.CharField(blank=True, help_text='Referral token', max_length=256, null=True, unique=True)),
                ('referral_link', models.CharField(blank=True, help_text='Referral link', max_length=256, null=True, unique=True)),
                ('link_clicked', models.IntegerField(default=0)),
                ('min_withdrawal_balance', models.DecimalField(decimal_places=2, default=0.0, help_text='Custom minimum balance required to withdraw earnings', max_digits=10)),
                ('active_payout_method', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='referrals.payoutmethod')),
                ('user', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='promoter', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='PromoterPayout',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True)),
                ('amount', models.IntegerField()),
                ('payout_method', models.CharField(choices=[('wise', 'Wise'), ('crypto', 'Crypto')], max_length=10)),
                ('tx_signature', models.CharField(blank=True, max_length=255, null=True)),
                ('promoter', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='promoter_payouts', to='referrals.promoter')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Referral',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True)),
                ('invitation_method', models.CharField(choices=[('email', 'Email'), ('link', 'Link')], max_length=10)),
                ('status', models.CharField(choices=[('signup', 'Signup'), ('active', 'Active'), ('refund', 'Refund')], max_length=10)),
                ('commission_rate', models.DecimalField(decimal_places=2, default=0.0, help_text='Commission rate at the moment of creating the referral', max_digits=5)),
                ('promoter', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='referrals', to='referrals.promoter')),
                ('user', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='referral', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='PromoterCommission',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True)),
                ('amount', models.IntegerField(null=True)),
                ('status', models.CharField(choices=[('pending', 'Pending'), ('paid', 'Paid'), ('failed', 'Failed'), ('refund', 'Refund')], default='pending', max_length=10)),
                ('failure_reason', models.TextField(blank=True, null=True)),
                ('invoice_external_id', models.CharField(blank=True, help_text='Chargebee invoice ID', max_length=255, null=True)),
                ('promoter', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='promoter_commission', to='referrals.promoter')),
                ('referral', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='referral_commission', to='referrals.referral')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



__all__ = [
    'referral_repository',
    'promoter_repository',
    'promoter_commission_repository',
    'promoter_payout_repository',
]

from .promoter_commission_repository import promoter_commission_repository
from .promoter_payout_repository import promoter_payout_repository
from .promoter_repository import promoter_repository
from .referral_repository import referral_repository



from io import StringIO

import pandas as pd


def parse_df_to_csv_string_without_index_col(df: pd.DataFrame) -> str:
    with StringIO() as buffer:
        df.to_csv(buffer, index=False)
        return buffer.getvalue()



import logging

from .base_repository import BaseRepository
from referrals.models import Promoter, PromoterPayout

logger = logging.getLogger(__name__)


class PromoterPayoutRepository(BaseRepository):
    def create_payout(self, promoter: Promoter, amount: float, payout_method, tx_signature: str = None):
        self.create(
            promoter=promoter,
            amount=amount,
            payout_method=payout_method,
            tx_signature=tx_signature,
        )


promoter_payout_repository = PromoterPayoutRepository(model=PromoterPayout)



import logging
import math
from decimal import Decimal
from typing import Optional

import pandas as pd
from pydantic import BaseModel

from referrals.choices import PromoterCommissionStatusChoices
from referrals.exceptions import ViewException
from referrals.helpers import parse_df_to_csv_string_without_index_col
from referrals.models import Promoter, PromoterPayout, PromoterCommission, Referral
from referrals.repositories import promoter_repository, promoter_payout_repository, promoter_commission_repository, \
    referral_repository

logger = logging.getLogger(__name__)


class PromoterPayoutDataRow(BaseModel):
    name: str
    recipientEmail: str
    amount: float
    sourceCurrency: str = "USD"
    targetCurrency: str = "USD"
    amountCurrency: str = "target"
    type: str = "EMAIL"


class PromoterPayoutService:
    def send_wise_csv_for_promoters_payouts(self, **kwargs) -> Optional[str]:
        """
        Generates a CSV file for Wise payouts and processes payouts for eligible promoters.

        This method retrieves all promoters eligible for Wise payouts, generates the necessary
        payout data, and creates payouts for those promoters whose current balance meets or exceeds
        the minimum withdrawal balance. The resulting data is converted into a CSV format string.

        Args:
            **kwargs: Additional keyword arguments to pass to the `PromoterPayoutDataRow`.

        Returns:
            Optional[str]: A CSV formatted string containing payout data, or None if no data is available.
        """
        promoters = promoter_repository.get_wise_payout_promoters()

        data = []
        for promoter in promoters:
            if promoter.current_balance > 0 and promoter.current_balance >= promoter.min_withdrawal_balance:
                payout_data_row = PromoterPayoutDataRow(
                    name=promoter.user.get_full_name(),
                    recipientEmail=promoter.active_payout_method.payment_address,
                    amount=promoter.current_balance,
                    **kwargs,
                )

                data.append(payout_data_row.model_dump())
                promoter_payout_repository.create_payout(
                    promoter, promoter.current_balance, payout_method='wise'
                )
                promoter_commission_repository.mark_commission_paid(promoter)
        if data:
            df = pd.DataFrame(data)
            return parse_df_to_csv_string_without_index_col(df)

    def calculate_commission(self, user_id: int,
                             amount_paid: int,
                             invoice_external_id: Optional[int] = None) -> Optional[PromoterCommission]:
        """
        Calculates and creates a commission for a promoter based on the referral's payment.

        This method checks if a commission has already been received for the given referral.
        If not, it calculates the commission amount and creates a new `PromoterCommission` entry.

        Args:
            user_id (int): The ID of the user who made the payment.
            amount_paid (int): The amount paid by the user in cents.
            invoice_external_id (Optional[int]): An optional external invoice ID.

        Returns:
            Optional[PromoterCommission]: The created commission, or None if no commission was created.
        """
        referral = referral_repository.get_referral_by_user_id(user_id)
        if not referral:
            return

        has_received_commission = promoter_repository.check_promoter_get_commission_from_referral(
            referral.promoter, referral
        )
        if has_received_commission:
            logger.info(f"Referrer {referral.promoter_id} already received commission from referral {referral.id}")
            return

        commission = self.create_commission(referral, amount_paid, invoice_external_id)
        if commission:
            logger.info(
                f"Commission {commission.id} created for promoter {referral.promoter_id} from referral {referral.id}"
            )
        return commission

    def create_commission(self, referral: Referral,
                          amount_paid: int,
                          invoice_external_id: Optional[int] = None) -> Optional[PromoterCommission]:
        """
        Creates a new commission entry for a promoter based on a referral's payment.

        Args:
            referral (Referral): The referral for which the commission is being created.
            amount_paid (int): The amount paid by the user in cents.
            invoice_external_id (Optional[int]): An optional external invoice ID.

        Returns:
            Optional[PromoterCommission]: The created commission, or None if no commission was created.
        """
        commission_amount = self.calculate_commission_amount(amount_paid, referral.commission_rate)

        commission = PromoterCommission(
            promoter=referral.promoter,
            referral=referral,
            amount=commission_amount,
            invoice_external_id=invoice_external_id,
        )
        commission.save()
        return commission

    @staticmethod
    def calculate_commission_amount(amount_paid: int, referral_commission_rate: Decimal) -> float:
        """
        Calculates the commission amount based on the payment amount and referral commission rate.

        Args:
            amount_paid (int): The amount paid by the user in cents.
            referral_commission_rate (Decimal): The commission rate associated with the referral.

        Returns:
            float: The calculated commission amount.
        """
        commission_rate = referral_commission_rate / 100
        price = amount_paid / 100

        return math.floor(Decimal(price) * commission_rate)

    @staticmethod
    def create_payout(promoter: Promoter, amount: float, payout_method):
        """
        Creates a payout record for a promoter and marks their pending commissions as paid.

        Args:
            promoter (Promoter): The promoter receiving the payout.
            amount (float): The payout amount.
            payout_method (str): The method used for the payout (e.g., 'wise', 'crypto').

        Returns:
            None
        """
        PromoterPayout.objects.create(
            promoter=promoter,
            amount=amount,
            payout_method=payout_method,
        )
        PromoterCommission.objects.filter(promoter=promoter, status="pending").update(status="paid")

    @staticmethod
    def calculate_refund(referral: Referral, amount_refunded: int, amount_paid: int,
                         invoice_external_id: Optional[int] = None) -> PromoterCommission:
        """
        Calculates the refund amount for a promoter's commission and creates a refund record.

        Args:
            referral (Referral): The referral associated with the refund.
            amount_refunded (int): The amount refunded in cents.
            amount_paid (int): The original amount paid in cents.
            invoice_external_id (Optional[int]): An optional external invoice ID.

        Returns:
            PromoterCommission: The created refund commission entry.

        Raises:
            ViewException: If no positive commission is found for the referral.
        """
        referral_commission = promoter_commission_repository.get_referral_positive_commission(referral)
        if not referral_commission:
            raise ViewException(
                f"No commission found for referral with id {referral.id}.",
                status_code=404
            )

        commission_paid = referral_commission.amount
        commission_refund_amount = -math.floor(commission_paid * amount_refunded / amount_paid)

        commission = PromoterCommission(
            promoter=referral.promoter,
            referral=referral,
            amount=commission_refund_amount,
            status=PromoterCommissionStatusChoices.REFUND,
            invoice_external_id=invoice_external_id,
        )
        commission.save()
        return commission


promoter_payout_service = PromoterPayoutService()



# Generated by Django 5.1 on 2024-08-24 14:32

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('referrals', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='payoutmethod',
            name='method',
            field=models.CharField(help_text='Payout method (e.g., wise, crypto, etc.)', max_length=20),
        ),
        migrations.AlterField(
            model_name='promotercommission',
            name='invoice_external_id',
            field=models.CharField(blank=True, help_text='e.g. Chargebee invoice ID', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='promoterpayout',
            name='payout_method',
            field=models.CharField(help_text='Payout method (e.g., wise, crypto, etc.)', max_length=20),
        ),
    ]



import logging

from referrals.choices import PromoterCommissionStatusChoices
from referrals.models import Promoter, Referral, PromoterCommission
from referrals.repositories.base_repository import BaseRepository

logger = logging.getLogger(__name__)


class PromoterCommissionRepository(BaseRepository):

    def mark_commission_paid(self, promoter: Promoter):
        self.filter(
            promoter=promoter,
            status__in=[PromoterCommissionStatusChoices.PENDING.value, PromoterCommissionStatusChoices.FAILED.value],
        ).update(status=PromoterCommissionStatusChoices.PAID.value)

    def mark_commission_failed_with_reason(self, promoter: Promoter, failure_reason: str):
        self.filter(promoter=promoter, status=PromoterCommissionStatusChoices.PENDING.value).update(
            status=PromoterCommissionStatusChoices.FAILED.value, failure_reason=failure_reason
        )

    def get_referral_positive_commission(self, referral: Referral):
        query = self.filter(
            referral=referral,
            status__in=[PromoterCommissionStatusChoices.PENDING, PromoterCommissionStatusChoices.PAID]
        )
        return query.first()


promoter_commission_repository = PromoterCommissionRepository(model=PromoterCommission)



from typing import Any, Dict
from urllib.parse import urlencode, urlparse, urlunparse, parse_qs


def append_query_params(url: str, params: Dict[str, Any]) -> str:
    """
    Appends query parameters to a given URL.

    Args:
        url (str): The base URL to which query parameters will be added.
        params (Dict[str, Any]): A dictionary of query parameters to add to the URL.

    Returns:
        str: The modified URL with the new query parameters appended.
    """
    url_parts = list(urlparse(url))
    query = dict(parse_qs(url_parts[4]))
    for key, value in params.items():
        if key in query:
            if isinstance(query[key], list):
                query[key].extend([value] if not isinstance(value, list) else value)
            else:
                query[key] = [query[key]] + ([value] if not isinstance(value, list) else value)
        else:
            query[key] = [value] if not isinstance(value, list) else value

    url_parts[4] = urlencode(query, doseq=True)
    return urlunparse(url_parts)



from decimal import Decimal

from django.contrib.auth.models import User
from django.core.validators import MinValueValidator
from django.db import models
from django.db import transaction
from django.db.models import Sum
from django.utils import timezone
from django.utils.functional import cached_property

from referrals.choices import InvitationMethodChoices, ReferralStateChoices, \
    PromoterCommissionStatusChoices


class TimeStampedModel(models.Model):
    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)

    class Meta:
        abstract = True

    def save(self, *args, **kwargs):
        self.updated = timezone.now()
        super().save(*args, **kwargs)


class ReferralProgram(TimeStampedModel):
    name = models.CharField(max_length=255, unique=True)
    commission_rate = models.DecimalField(
        max_digits=5,
        decimal_places=2,
        help_text="Commission rate as a percentage",
        validators=[MinValueValidator(Decimal("0.01"))],
    )
    is_active = models.BooleanField(
        default=True, help_text="Indicates if this is the currently active referral program"
    )
    min_withdrawal_balance = models.DecimalField(
        max_digits=10, decimal_places=2, default=0.00, help_text="Minimum balance required to withdraw earnings"
    )

    @transaction.atomic
    def save(self, *args, **kwargs):
        if self.is_active:
            ReferralProgram.objects.filter(is_active=True).update(is_active=False)
            Promoter.objects.filter(min_withdrawal_balance__lt=self.min_withdrawal_balance).update(
                min_withdrawal_balance=self.min_withdrawal_balance
            )

        super(ReferralProgram, self).save(*args, **kwargs)

    @classmethod
    def get_active_referral_program(cls):
        return cls.objects.filter(is_active=True).first()


class PayoutMethod(models.Model):
    method = models.CharField(max_length=20, help_text="Payout method (e.g., wise, crypto, etc.)")
    payment_address = models.CharField(max_length=100, help_text="Payment address (email or solana wallet)")

    def __str__(self):
        return self.method


class Promoter(TimeStampedModel):
    user = models.OneToOneField(User, on_delete=models.CASCADE, related_name="promoter")
    referral_token = models.CharField(max_length=256, blank=True, null=True, unique=True, help_text="Referral token")
    referral_link = models.CharField(max_length=256, blank=True, null=True, unique=True, help_text="Referral link")
    active_payout_method = models.ForeignKey(PayoutMethod, on_delete=models.SET_NULL, null=True, blank=True)
    link_clicked = models.IntegerField(default=0)
    min_withdrawal_balance = models.DecimalField(
        max_digits=10,
        decimal_places=2,
        default=0.00,
        help_text="Custom minimum balance required to withdraw earnings",
    )

    @cached_property
    def total_earned(self) -> int:
        return self.promoter_commission.aggregate(total=Sum("amount"))["total"] or 0

    @cached_property
    def total_paid(self) -> int:
        return (
                self.promoter_payouts.aggregate(
                    total=Sum("amount")
                )["total"]
                or 0
        )

    @cached_property
    def current_balance(self) -> int:
        return self.total_earned - self.total_paid

    def __str__(self):
        return f"{self.user.email} - {self.referral_link}"

    def save(self, *args, **kwargs):
        if self.pk is None:
            active_program = ReferralProgram.get_active_referral_program()
            if active_program:
                self.min_withdrawal_balance = active_program.min_withdrawal_balance
        super(Promoter, self).save(*args, **kwargs)


class Referral(TimeStampedModel):
    user = models.OneToOneField(User, on_delete=models.CASCADE, related_name="referral")
    promoter = models.ForeignKey(Promoter, related_name="referrals", on_delete=models.CASCADE)
    invitation_method = models.CharField(max_length=10, choices=InvitationMethodChoices.choices)
    status = models.CharField(max_length=10, choices=ReferralStateChoices.choices)
    commission_rate = models.DecimalField(
        max_digits=5, decimal_places=2, help_text="Commission rate at the moment of creating the referral", default=0.00
    )

    def save(self, *args, **kwargs):
        if self.pk is None:
            active_program = ReferralProgram.get_active_referral_program()
            if active_program:
                self.commission_rate = active_program.commission_rate
        super().save(*args, **kwargs)


class PromoterCommission(TimeStampedModel):
    promoter = models.ForeignKey(Promoter, related_name="promoter_commission", on_delete=models.CASCADE)
    referral = models.ForeignKey(Referral, related_name="referral_commission", on_delete=models.CASCADE)
    amount = models.IntegerField(null=True)
    status = models.CharField(
        max_length=10, choices=PromoterCommissionStatusChoices.choices, default=PromoterCommissionStatusChoices.PENDING
    )
    failure_reason = models.TextField(null=True, blank=True)
    invoice_external_id = models.CharField(null=True, max_length=255, blank=True, help_text="e.g. Chargebee invoice ID")


class PromoterPayout(TimeStampedModel):
    promoter = models.ForeignKey(Promoter, related_name="promoter_payouts", on_delete=models.CASCADE)
    amount = models.IntegerField()
    payout_method = models.CharField(max_length=20, null=False, help_text="Payout method (e.g., wise, crypto, etc.)")
    tx_signature = models.CharField(max_length=255, null=True, blank=True)
